import copy
import ipaddress
import socket
import typing as t

from flask import current_app, url_for, g
from sqlalchemy import or_

from dm.utils.typos import ScalarListType, Gate as TGate
from dm.web import db, errors
from .base import UUIDistributedEntityMixin
from .gate import Gate
from .route import Route
from ... import defaults
# TODO: handle multiple networks (IP gateways) on a server with netifaces
from ...utils.helpers import get_ips_listening_for


class Server(db.Model, UUIDistributedEntityMixin):
    __tablename__ = 'D_server'
    order = 10

    name = db.Column(db.String(255), nullable=False, unique=True)
    granules = db.Column(ScalarListType())
    _me = db.Column("me", db.Boolean, default=False)

    route = db.relationship("Route", primaryjoin="Route.destination_id==Server.id", uselist=False,
                            back_populates="destination", cascade="all, delete-orphan")
    gates = db.relationship("Gate", back_populates="server", cascade="all, delete-orphan")

    # software_list = db.relationship("SoftwareServerAssociation", back_populates="server")

    def __init__(self, name: str, granules: t.List[str] = None,
                 dns_or_ip: t.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address] = None, port: int = None,
                 gates: t.List[t.Union[TGate, t.Dict[str, t.Any]]] = None, me: bool = False, **kwargs):
        UUIDistributedEntityMixin.__init__(self, **kwargs)
        self.name = name
        if port or dns_or_ip:
            self.add_new_gate(dns_or_ip or self.name, port or defaults.DEFAULT_PORT)

        if gates:
            for gate in gates:
                if isinstance(gate, tuple):
                    self.add_new_gate(*gate)
                elif isinstance(gate, dict):
                    gate['server'] = self
                    Gate.from_json(gate)

        assert 'all' not in (granules or [])
        self.granules = granules or []
        self._me = me
        # create an empty route
        if not me:
            Route(self)

    @property
    def external_gates(self):
        e_g = []
        for g in self.gates:
            if not g.ip:
                try:
                    ip = ipaddress.ip_address(
                        socket.getaddrinfo(g.dns, 0, family=socket.AF_INET, proto=socket.IPPROTO_TCP)[0][4][0])
                except socket.gaierror:
                    e_g.append(g)
                    continue
                except KeyError:
                    e_g.append(g)
                    continue
            else:
                ip = g.ip
            if not ip.is_loopback:
                e_g.append(g)
        return e_g

    @property
    def localhost_gates(self):
        l_g = []
        for g in self.gates:
            if not g.ip:
                try:
                    ip = ipaddress.ip_address(
                        socket.getaddrinfo(g.dns, 0, family=socket.AF_INET, proto=socket.IPPROTO_TCP)[0][4][0])
                except socket.gaierror:
                    continue
                except KeyError:
                    continue
            else:
                ip = g.ip
            if ip.is_loopback:
                l_g.append(g)
        return l_g

    @property
    def hidden_gates(self):
        hg = [g for g in self.gates if g.hidden]
        hg.sort(key=lambda x: x.last_modified_at or get_now())
        return hg

    def add_new_gate(self, dns_or_ip: t.Union[str, ipaddress.IPv4Address, ipaddress.IPv6Address], port: int,
                     hidden=None):
        ip = None
        dns = None
        if dns_or_ip:
            try:
                ip = ipaddress.ip_address(dns_or_ip)
            except ValueError:
                dns = dns_or_ip
        return Gate(server=self, port=port, dns=dns, ip=ip, hidden=hidden)

    def __str__(self, ):
        return f"{self.name}"

    def url(self, view: str = None, **values) -> str:
        """
        generates the full url to access the server. Uses url_for to generate the full_path.

        Parameters
        ----------
        view
        values

        Raises
        -------
        ConnectionError:
            if server is unreachable
        """
        scheme = current_app.config['PREFERRED_URL_SCHEME'] or 'https'
        gate = None
        # route = Route.query.filter_by(destination_id=self.id)
        route = self.route
        if self._me and (route is None or route.cost is None):
            try:
                gate = self.localhost_gates[0]
            except IndexError:
                current_app.logger.warning(
                    f"No localhost set for '{self}'. Trying connection through another gate")
                if len(self.gates) == 0:
                    raise RuntimeError(f"No gate set for server '{self}'")
                gate = self.gates[0]
        elif route is not None and route.cost == 0:
            gate = route.gate
        else:
            if route.proxy_server:
                gate = route.proxy_server.route.gate

        if not gate:
            raise errors.UnreachableDestination(self, getattr(g, 'server', None))

        root_path = f"{scheme}://{gate}"

        if view is None:
            return root_path
        else:
            with current_app.test_request_context():
                return root_path + url_for(view, **values)

    @classmethod
    def get_neighbours(cls) -> t.List['Server']:
        return db.session.query(cls).join(cls.route).filter(Route.cost == 0).all()

    @classmethod
    def get_not_neighbours(cls) -> t.List['Server']:
        return db.session.query(cls).outerjoin(cls.route).filter(
            or_(or_(Route.cost > 0, Route.cost == None), cls.route == None)).filter(Server._me == False).all()

    def to_json(self, add_gates=False, human=False):
        data = super().to_json()
        data.update(
            {'name': self.name, 'granules': self.granules})
        if add_gates:
            data.update(gates=[])
            for g in self.gates:
                json_gate = g.to_json(human=human)
                json_gate.pop('server_id', None)
                json_gate.pop('server', None) # added to remove when human set
                data['gates'].append(json_gate)
        return data

    @classmethod
    def from_json(cls, kwargs) -> 'Server':
        kwargs = copy.deepcopy(kwargs)
        gates = kwargs.pop('gates', [])
        server = super().from_json(kwargs)
        for gate in gates:
            gate.update(server=server)
            Gate.from_json(gate)
        return server

    @classmethod
    def get_current(cls) -> 'Server':
        return db.session.query(cls).filter_by(_me=True).one()

    @staticmethod
    def set_initial():
        server = Server.query.filter_by(_me=True).all()
        if len(server) == 0:
            gates = get_ips_listening_for()
            server_name = current_app.config.get('SERVER_NAME') or defaults.HOSTNAME

            server = Server(name=server_name,
                            gates=gates,
                            me=True)
            db.session.add(server)
        elif len(server) > 1:
            raise ValueError('Multiple servers found as me.')



import threading
from datetime import datetime

from flask import request, current_app, g, jsonify
from flask_jwt_extended import jwt_required

from dm import defaults
from dm.domain.entities import Catalog
from dm.domain.entities.locker import Scope, State, Locker
from dm.web import db, errors
from dm.web.api_1_0 import api_bp
from dm.web.decorators import securizer, forward_or_dispatch, validate_schema
from dm.web.json_schemas import locker_prevent_post, locker_unlock_lock_post


@api_bp.route('/locker', methods=['GET'])
@forward_or_dispatch
@jwt_required
@securizer
def locker():
    data = []
    for l in Locker.query.all():
        data.append({l.scope.name: l.state.name})
    return jsonify(data), 200


def revert_preventing(app, scope, applicant):
    with app.app_context():
        l = Locker.query.with_for_update().get(scope)
        if l.state == State.PREVENTING and l.applicant == applicant:
            l.state = State.UNLOCKED
            l.applicant = None
        db.session.commit()


@api_bp.route('/locker/prevent', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(POST=locker_prevent_post)
def locker_prevent():
    json_data = request.get_json()
    l: Locker = Locker.query.with_for_update().get(Scope[json_data['scope']])
    current_app.logger.debug(f"PreventLock requested on {json_data.get('scope')} from {g.source}")

    # when orchestration scope check if applicant is the same as the current
    if Scope[json_data['scope']] == Scope.ORCHESTRATION \
            and l.state in (State.PREVENTING, State.LOCKED) \
            and l.applicant == json_data.get('applicant'):
        return {'message': f"{Scope[json_data['scope']].name} already in {l.state.name} state"}, 210

    # check status from current scope
    if l.state == State.UNLOCKED:
        # check priority
        prioritary_lockers = Locker.query.filter(Locker.scope != l.scope).all()
        prioritary_lockers = [pl for pl in prioritary_lockers if pl.scope < l.scope]
        cond = any([pl.state in (State.PREVENTING, State.LOCKED) for pl in prioritary_lockers])
        if not cond:
            # catalog serialization
            if json_data['scope'] != Scope.UPGRADE.name:
                datemark = datetime.strptime(json_data['datemark'], defaults.DATEMARK_FORMAT)
                catalog_ver = Catalog.max_catalog()
                if datemark < catalog_ver:
                    raise errors.ObsoleteCatalog(catalog_ver, datemark)
            l.state = State.PREVENTING
            l.applicant = json_data.get('applicant')
            th = threading.Timer(defaults.TIMEOUT_PREVENTING_LOCK, revert_preventing,
                                 (current_app._get_current_object(), l.scope, l.applicant))
            th.daemon = True
            db.session.commit()
            th.start()
            return {'message': 'Preventing lock acquired'}, 200
        else:
            raise errors.PriorityLocker(l.scope)
    else:
        raise errors.StatusLockerError(l.scope, 'P', l.state)


@api_bp.route('/locker/lock', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(POST=locker_unlock_lock_post)
def locker_lock():
    json_data = request.get_json()
    l: Locker = Locker.query.with_for_update().get(Scope[json_data['scope']])
    current_app.logger.debug(f"Lock requested on {json_data.get('scope')} from {g.source}")

    if Scope[json_data['scope']] == Scope.ORCHESTRATION \
            and l.state == State.LOCKED \
            and l.applicant == json_data.get('applicant'):
        return {'message': f"{Scope[json_data['scope']]} already in {l.state} state"}, 210

    if l.state == State.PREVENTING:
        if l.applicant == json_data['applicant']:
            l.state = State.LOCKED
            db.session.commit()
            current_app.logger.debug(f"Lock from {g.source} on {l.scope.name} acquired")
            return {'message': 'Locked'}, 200
        else:
            raise errors.ApplicantLockerError(l.scope)
    else:
        raise errors.StatusLockerError(l.scope, 'L', l.state)


@api_bp.route('/locker/unlock', methods=['POST'])
@forward_or_dispatch
@jwt_required
@securizer
@validate_schema(POST=locker_unlock_lock_post)
def locker_unlock():
    json_data = request.get_json()
    l: Locker = Locker.query.with_for_update().get(Scope[json_data['scope']])
    current_app.logger.debug(f"Unlock requested on {json_data.get('scope')} from {g.source}")

    if Scope[json_data['scope']] == Scope.ORCHESTRATION and l.state == State.UNLOCKED:
        return {'message': f"{Scope[json_data['scope']]} already in {l.state} state"}, 210

    if l.state == State.PREVENTING or l.state == State.LOCKED:
        if l.applicant == json_data['applicant']:
            l.state = State.UNLOCKED
            l.applicant = None
            db.session.commit()
            current_app.logger.debug(f"Lock on {l.scope.name} released")
            return {'message': 'UnLocked'}, 200
        else:
            raise errors.ApplicantLockerError(l.scope)
    else:
        raise errors.StatusLockerError(l.scope, 'U', l.state)



import threading
import time
import typing as t
from collections import namedtuple

from dm.utils.typos import Id


class Event(object):

    def __init__(self, id_, data):
        self._id = id_
        self.data = data

    @property
    def id(self):
        return self._id


_RegistryContainer = namedtuple('_RegistryContainer', ['func', 'args', 'kwargs', 'birth'])


class EventHandler(object):

    def __init__(self, discard_after=3600):
        self.discard_after = discard_after
        self._registry: t.Dict[Id, _RegistryContainer] = {}
        self._pending_events: t.Dict[Id, t.Tuple[Event, float]] = {}
        self._lock = threading.Lock()

    def discard(self):
        now = time.time()
        for k, c in dict(self._registry).items():
            if now - c.birth >= self.discard_after:
                self._registry.pop(k)
        for k, v in dict(self._pending_events).items():
            if now - v[1] >= self.discard_after:
                self._pending_events.pop(k)

    def register(self, key, func: t.Callable[..., None], args=None, kwargs=None):
        event = None
        with self._lock:
            if key not in self._registry and key not in self._pending_events:
                self._registry[key] = _RegistryContainer(func, args or (), kwargs or {}, time.time())
            elif key in self._pending_events:
                event, t = self._pending_events.pop(key)
            else:
                raise ValueError('event ID duplicated')
            self.discard()
        if event:
            func(event, *(args or ()), **(kwargs or {}))

    def dispatch(self, event: Event):
        func = None
        with self._lock:
            try:
                func, args, kwargs, birth = self._registry.pop(event.id)
            except KeyError:
                self._pending_events[event.id] = (event, time.time())
            self.discard()
        if func:
            func(event, *args, **kwargs)



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Step
from dm.web import db
from dm.web.decorators import forward_or_dispatch, securizer, validate_schema, lock_catalog
from dm.web.json_schemas import step_parents


class StepRelationshipParents(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        return dict(parent_step_ids=[str(ps.id) for ps in s.parents]), 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_parents)
    @lock_catalog
    def patch(self, step_id):
        s: Step = Step.query.get_or_404(step_id)
        parent_step_ids = request.get_json()['parent_step_ids']
        parent_steps = []
        for parent_step_id in parent_step_ids:
            parent_steps.append(Step.query.get_or_404(parent_step_id))
        s.orchestration.set_parents(s, parent_steps)

        db.session.commit()

        return dict(parent_step_ids=[str(ps.id) for ps in s.parents]), 200


    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_parents)
    @lock_catalog
    def post(self, step_id):
        s = Step.query.get_or_404(step_id)
        parent_step_ids = request.get_json()['parent_step_ids']
        parent_steps = []
        for parent_step_id in parent_step_ids:
            parent_steps.append(Step.query.get_or_404(parent_step_id))
        s.orchestration.add_parents(s, parent_steps)

        db.session.commit()

        return dict(parent_step_ids=[str(ps.id) for ps in s.parents]), 200

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(step_parents)
    @lock_catalog
    def delete(self, step_id):
        s = Step.query.get_or_404(step_id)
        parent_step_ids = request.get_json()['parent_step_ids']
        parent_steps = []
        for parent_step_id in parent_step_ids:
            parent_steps.append(Step.query.get_or_404(parent_step_id))

        s.orchestration.delete_parents(s, parent_steps)

        db.session.commit()

        return dict(parent_step_ids=[str(ps.id) for ps in s.parents]), 200


import argparse as argparse
import json
import uuid
from pprint import pprint

import pygments
import requests
from prompt_toolkit import print_formatted_text
from prompt_toolkit.completion import merge_completers
from prompt_toolkit.formatted_text import PygmentsTokens
from pygments.lexers.web import JSONLexer

from dm import defaults
from dm.domain.entities import ActionType
from dm.dshell import environ
from dm.dshell.completer import *
from dm.dshell.helpers import name2id
from dm.dshell.prompts.action_template import subprompt as action_prompt
from dm.dshell.prompts.orchestration import subprompt as orch_prompt
from dm.dshell.utils import clean_none
from dm.utils.helpers import get_now


def status(node):
    for node_id in node:
        try:
            uuid.UUID(node_id)
        except Exception:
            node_id = name2id('api_1_0.serverlist', node_id)
        resp = ntwrk.get('root.healthcheck', headers={'D-Destination': node_id})
        pprint(resp.msg if resp.code else str(resp.exception))


def ping(node):
    for node_id in node:
        try:
            uuid.UUID(node_id)
        except Exception:
            node_id = name2id('api_1_0.serverlist', node_id)
        resp = ntwrk.post('root.ping', headers={'D-Destination': node_id},
                          json={'start_time': get_now().strftime(defaults.DATETIME_FORMAT)})
        pprint(resp.msg if resp.code else str(resp.exception))


def server_list(name=None, iden=None, detail=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if name is not None:
        view_data.update({'filter[name]': name})
    if iden:
        view_data.update({'filter[id]': iden})
    if detail:
        view_data.update(params='gates')
    resp = ntwrk.get('api_1_0.serverlist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        pprint(filtered_data)
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def orch_list(**params):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if params.get('id', None):
        view_data.update({'filter[id]': params['id']})
    if params.get('name'):
        view_data.update({'filter[name]': params['name']})
    if params.get('version', None):
        view_data.update({'filter[version]': params['version']})
    if params.get('details', None):
        view_data.update({'params': ['steps', 'vars', 'target', 'action', 'human']})
    else:
        view_data.update({'params': ['vars', 'target', 'human']})
    resp = ntwrk.get('api_1_0.orchestrationlist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if params.get('like', None):
            for server in data:
                if params.get('like') in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        pprint(filtered_data)
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def orch_create(**params):
    orch_prompt({'name': params['name'], 'steps': []},
                changed=True,
                ask_all=params['prompt'], parent_prompt='Δ ')


def orch_copy(**params):
    resp = ntwrk.get('api_1_0.orchestrationresource',
                     view_data={'orchestration_id': params.get('orchestration_id')})

    if resp.ok:
        orch = resp.msg
        orch.pop('target', None)
        orch.pop('params', None)
        orch.pop('id', None)
        orch.pop('version', None)
        orch.pop('last_modified_at', None)
        resp = ntwrk.get('api_1_0.steplist',
                         view_data={'filter[orchestration_id]': params.get('orchestration_id')})
        if resp.ok:
            orch['steps'] = resp.msg

            # remove unwanted parameters
            for s in orch['steps']:
                s.pop('last_modified_at', None)
                s.pop('created_on', None)

            orch_prompt(entity=orch, parent_prompt='Δ ')


def orch_load(file):
    try:
        orch = json.load(file)
    except Exception as e:
        print(e)
    else:
        orch_prompt(orch, parent_prompt='Δ ')


def orch_run(orchestration_id, **params):
    if not params['hosts']:
        print('No target specified')
        return
    resp = ntwrk.post('api_1_0.launch_orchestration',
                      view_data={'orchestration_id': orchestration_id, 'params': 'human'}, json=params)
    if resp.code is not None:
        pprint(resp.msg)
    else:
        pprint(str(resp.exception))


def action_list(iden=None, name=None, version=None, like: str = None, last: int = None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if iden:
        view_data.update({'filter[id]': iden})
    if name:
        view_data.update({'filter[name]': name})
    if version:
        view_data.update({'filter[version]': version})
    resp = ntwrk.get('api_1_0.actiontemplatelist', view_data=view_data, **kwargs)
    if resp.code == 200:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        tokens = list(pygments.lex(json.dumps(filtered_data[:last or len(filtered_data)], indent=2), lexer=JSONLexer()))
        print_formatted_text(PygmentsTokens(tokens))
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def action_create(name, action_type, prompt):
    action_prompt({'name': name, 'action_type': action_type}, changed=True, ask_all=prompt, parent_prompt='Δ ')


#
# def action_modify(**params):
#     resp = ntwrk.get('api_1_0.actiontemplateresource',
#                      view_data={'action_template_id': params.get('action_template_id')})
#     if resp.code == 200:
#         action_prompt(entity=resp.msg, parent_prompt='Δ ')


def action_load(file):
    try:
        action = json.load(file)
    except Exception as e:
        print(e)
    else:
        action_prompt(action, parent_prompt='Δ ')


def software_add(name, version, file, family=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(name=name, version=version, file=file)
    if family is not None:
        json_data.update(family=family)
    resp = ntwrk.post(view='api_1_0.softwarelist', json=json_data, **kwargs)
    if resp.code == 200:
        pprint(resp.msg)
    else:
        pprint(resp.msg if resp.code is not None else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def software_send(dest_server_id, software_id=None, software=None, version=None, file=None, dest_path=None,
                  background=True, force=False):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(dest_server_id=dest_server_id, dest_path=dest_path, force=force)
    if software_id:
        json_data.update(software_id=software_id)
    elif file:
        json_data.update(file=file)
    else:
        json_data.update(software=software, version=version)

    if not background:
        json_data.update(include_transfer_data=True)
    resp = ntwrk.post('api_1_0.send', json=json_data, **kwargs)
    pprint(resp.msg if resp.code is not None else str(resp.exception) if str(
        resp.exception) else resp.exception.__class__.__name__)


def software_list(name=None, version=None, detail=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if name:
        view_data.update({'filter[name]': name})
    if version:
        view_data.update({'filter[version]': version})
    if detail:
        view_data.update(params='servers')
    resp = ntwrk.get('api_1_0.softwarelist', view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        pprint(filtered_data)
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def transfer_list(iden=None, status=None, like=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    if iden:
        view_data.update({'filter[id]': iden})
    if status:
        view_data.update({'filter[status]': ','.join(status)})
    resp = ntwrk.get('api_1_0.transferlist', view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        filtered_data = []
        if like:
            for server in data:
                if like in server.get('name'):
                    filtered_data.append(server)
        else:
            filtered_data = data
        pprint(filtered_data)
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def transfer_cancel(transfer_id):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    data = {'status': 'CANCELED'}
    resp = ntwrk.patch('api_1_0.transferresource', view_data={'transfer_id': transfer_id}, json=data, **kwargs)
    pprint(resp.msg if resp.code else str(resp.exception) if str(
        resp.exception) else resp.exception.__class__.__name__)


def exec_list(orch=None, server=None, last=None, asc=None, detail=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    view_data = dict()
    view = 'api_1_0.orchexecutionlist'

    if orch:
        view_data.update({'filter[orchestration_id]': orch})

    if server:
        view_data.update({'filter[server_id]': server})

    # url_parameters
    if detail:
        view_data.update({'params': ['steps', 'human']})
    else:
        view_data.update({'params': ['human']})

    resp = ntwrk.get(view, view_data=view_data, **kwargs)
    if resp.ok:
        data = resp.msg or []
        # post process
        if asc:
            data.sort(key=lambda x: x.get('start_time'))
        else:
            data.sort(key=lambda x: x.get('start_time'), reverse=True)
        tokens = list(pygments.lex(json.dumps(data[:last or len(data)], indent=2), lexer=JSONLexer()))
        print_formatted_text(PygmentsTokens(tokens))
    else:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


def cmd(command, hosts, timeout=None, input=None):
    if isinstance(command, list):
        command = ' '.join(command)

    data = {'command': command, 'hosts': hosts}
    if timeout:
        data.update(timeout=timeout)
    if input:
        data.update(input=input.replace('\\n', '\n').replace('\\t', '\t'))
    resp = ntwrk.post('api_1_0.launch_command', view_data={'params': 'human'}, json=data)
    if resp.code:
        tokens = list(pygments.lex(json.dumps(resp.msg, indent=2), lexer=JSONLexer()))
        print_formatted_text(PygmentsTokens(tokens))
    else:
        pprint(str(resp.exception))


def logfed_list():
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    resp = ntwrk.get('api_1_0.loglist', view_data={'params': 'human'}, **kwargs)
    pprint(resp.msg if resp.code else str(resp.exception) if str(
        resp.exception) else resp.exception.__class__.__name__)


def logfed_subscribe(src_server_id, target, dest_server_id, include=None, exclude=None, dest_folder=None,
                     recursive=None, mode=None):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    json_data = dict(src_server_id=src_server_id, target=target, dest_server_id=dest_server_id, include=include,
                     exclude=exclude, dest_folder=dest_folder, recursive=recursive, mode=mode)
    json_data = clean_none(json_data)
    resp = ntwrk.post('api_1_0.loglist', json=json_data, **kwargs)
    pprint(resp.msg if resp.code else str(resp.exception) if str(
        resp.exception) else resp.exception.__class__.__name__)


def logfed_unsubscribe(log_id):
    kwargs = dict(verify=environ.get('SSL_VERIFY'))
    resp = ntwrk.delete('api_1_0.logresource', {'log_id': log_id}, **kwargs)
    if resp.code is not None and resp.msg or resp.exception is not None:
        pprint(resp.msg if resp.code else str(resp.exception) if str(
            resp.exception) else resp.exception.__class__.__name__)


# def locker_lock(**params):
#     pass
#
# def locker_unlock(**params):
#     if params['servers'] is None:
#         resp = get('api_1_0.serverlist')
#         resp.raise_if_not_ok()
#         servers = [s.id for s in resp.msg]
#     else:
#         servers = params['servers']
#
#     for server in servers:
#         resp = post('api_1_0.locker_unlock')

def login(username=None, password=None):
    try:
        ntwrk.login(username, password)
    except requests.exceptions.ConnectionError as e:
        print(f"Unable to contact with {environ.get('SCHEME')}://{environ.get('SERVER')}:{environ.get('PORT')}/")
    except Exception as e:
        pprint(e)


def logging_cmd(logger, level):
    logger = logging.getLogger(logger)
    if logger:
        logger.setLevel(level)
    else:
        print(f"Logger '{logger}' does not exist")


def env_list():
    for k, v in environ._environ.items():
        pprint(f"{k}={v}")


def env_get(key):
    pprint(f"{environ.get(key, None)}")


def env_set(key, value):
    environ.set(key, value)


nested_dict = {
    'status': [{'argument': 'node', 'nargs': '+', 'completer': server_completer}, status],
    'ping': [{'argument': 'node', 'nargs': '+', 'completer': server_completer}, ping],
    'server': {
        'list': [{'argument': '--detail', 'action': 'store_true'},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--name', 'completer': server_name_completer},
                  {'argument': '--id', 'dest': 'iden', 'completer': server_completer},
                  # {'argument': '--last', 'action': 'store', 'type': int}
                  ],
                 server_list
                 ],
    },
    'orch': {
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': orch_ver_completer},
                 {'argument': '--details', 'action': 'store_true'},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': orch_completer},
                  {'argument': '--name', 'completer': orch_name_completer}],
                 orch_list
                 ],
        'create': [{'argument': 'name'},
                   {'argument': '--prompt', 'action': "store_true",
                    'help': 'does not ask for every orch parameter one by one'},
                   orch_create],
        'copy': [{'argument': 'orchestration_id', 'completer': orch_completer},
                 orch_copy],
        'load': [{'argument': 'file', 'type': argparse.FileType('r')},
                 orch_load],
        'run': [{'argument': 'orchestration_id', 'completer': orch_completer},
                {'argument': '--target', 'action': DictAction, 'nargs': "+", 'dest': 'hosts',
                 'completer': merge_completers([server_completer, granule_completer])},
                {'argument': '--param', 'action': ParamAction, 'nargs': "+", 'dest': 'params', 'default': {}},
                {'argument': '--foreground', 'dest': 'background', 'action': 'store_false'},
                orch_run],
    },
    'action': {
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': action_ver_completer},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': action_completer},
                  {'argument': '--name', 'completer': action_name_completer}],
                 action_list
                 ],
        'create': [{'argument': 'name'},
                   {'argument': 'version', 'type': int},
                   {'argument': 'action_type', 'choices': [at.name for at in ActionType if at.name != 'NATIVE']},
                   {'argument': '--prompt', 'action': "store_true",
                    'help': 'does not ask for every action parameter one by one'},
                   action_create],
        # 'modify': [{'argument': 'action_template_id', 'completer': action_completer},
        #            action_modify],
        'load': [{'argument': 'file', 'type': argparse.FileType('r')},
                 action_load],
    },
    'exec': {
        'list': [{'argument': '--orch', 'completer': orch_completer},
                 {'argument': '--server', 'completer': server_completer},
                 {'argument': '--last', 'type': int},
                 {'argument': '--asc', 'action': 'store_true'},
                 {'argument': '--detail', 'action': 'store_true'},
                 exec_list]
    },
    'software': {
        'add': [{'argument': 'name'},
                {'argument': 'version'},
                {'argument': 'file'},
                {'argument': '--family', 'completer': software_family_completer},
                software_add],
        'list': [{'argument': '--version', 'action': 'store', 'type': int,
                  'completer': software_ver_completer},
                 {'argument': '--detail', 'action': 'store_true'},
                 [{'argument': '--json', 'action': 'store_true'},
                  {'argument': '--table', 'action': 'store_true'}],
                 [{'argument': '--like'},
                  {'argument': '--id', 'completer': software_completer},
                  {'argument': '--name', 'completer': software_name_completer}],
                 software_list
                 ],
        'send': [{'argument': 'software_id', 'completer': software_completer},
                 {'argument': 'dest_server_id', 'completer': server_completer},
                 {'argument': '--dest_path'},
                 {'argument': '--foreground', 'action': 'store_false', 'dest': 'background', 'default': True},
                 {'argument': '--force', 'action': 'store_true'},
                 software_send], },
    'transfer': {
        'cancel': [{'argument': 'transfer_id'},
                   transfer_cancel],
        'list': [
            {'argument': '--status', 'action': "extend", 'nargs': '+',
             'choices': ['WAITING_CHUNKS', 'IN_PROGRESS', 'COMPLETED', 'CHECKSUM_ERROR', 'SIZE_ERROR',
                         'CANCELED']},
            [{'argument': '--id'},
             {'argument': '--last', 'action': 'store', 'type': int}, ],
            transfer_list
        ]},
    'cmd': [{'argument': 'command', 'nargs': '+'},
            {'argument': '--target', 'action': 'extend', 'nargs': "+", 'dest': 'hosts',
             'completer': merge_completers([server_completer, granule_completer])},
            {'argument': '--timeout', 'type': int, 'help': 'timeout in seconds to wait for command to terminate'},
            {'argument': '--input'},
            cmd],
    'logfed': {'subscribe': {'log': [{'argument': 'src_server_id',
                                      'metavar': 'source_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': 'target',
                                      'metavar': 'file',
                                      'help': 'log file to watch out'},
                                     {'argument': 'dst_server_id',
                                      'metavar': 'destination_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': '--mode', 'choices': ['REPO_MIRROR', 'REPO_ROOT', 'MIRROR'],
                                      'help': 'defines where the log will be sent on destination. REPO_MIRROR send the '
                                              'file inside the dest LOG folder and mantains absolute path from origin. '
                                              'REPO_ROOT sends the file inside the dest LOG without mantaining origin'
                                              '\'s path. MIRROR preserves source path and tries to create at dest'},
                                     {'argument': '--dest_folder',
                                      'help': 'destination folder to send logs. If not specified, '
                                              'default mode REPO_MIRROR is used'},
                                     logfed_subscribe
                                     ],
                             'dir': [{'argument': 'target',
                                      'metavar': 'folder',
                                      'help': 'folder to watch out for files to send'},
                                     {'argument': 'src_server_id',
                                      'metavar': 'source_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': 'dst_server_id',
                                      'metavar': 'destination_server_id',
                                      'help': 'source server to get the logs from',
                                      'completer': server_completer},
                                     {'argument': '--include',
                                      'help': 'regular expression used to filter which '
                                              'files and folders should subscribe'},
                                     {'argument': '--exclude',
                                      'help': 'regular expression used to filter which files and folders '
                                              'should NOT subscribe'},
                                     {'argument': '--recursive',
                                      'help': 'enters in each folder to find files to send'},
                                     {'argument': '--mode', 'choices': ['REPO_MIRROR', 'REPO_ROOT', 'MIRROR'],
                                      'help': 'defines where the log will be sent on destination. REPO_MIRROR send the '
                                              'file inside the dest LOG folder and mantains absolute path from origin. '
                                              'REPO_ROOT sends the file inside the dest LOG without mantaining origin'
                                              '\'s path. MIRROR preserves source path and tries to create at dest'},
                                     {'argument': '--dest_folder',
                                      'help': 'destination folder to send logs. If not specified, '
                                              'default mode REPO_MIRROR is used'},
                                     logfed_subscribe]
                             },
               'unsubscribe': [{'argument': 'log_id', 'completer': logfed_completer},
                               logfed_unsubscribe],
               'list': [logfed_list],
               },
    # "locker": {"lock": [{'argument': "scope", 'choices': ["CATALOG", "ORCHESTRATION", "UPGRADE"]},
    #                     {'argument': "servers", 'nargs': '*', 'completer': server_completer},
    #                     locker_lock],
    #            "unlock": [{'argument': "scope", 'choices': ["CATALOG", "ORCHESTRATION", "UPGRADE"]},
    #                       {'argument': "servers", 'nargs': '*', 'completer': server_completer},
    #                       locker_unlock]},
    "login": [{'argument': 'username', 'nargs': '?'},
              login],
    "logging": [{'argument': 'logger', 'completer': logger_completer},
                {'argument': 'level', 'choices': ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']},
                logging_cmd],
    "env": {"list": [env_list],
            "get": [{'argument': 'key', 'completer': DshellWordCompleter(environ._environ.keys())},
                    env_get],
            "set": [{'argument': 'key', 'completer': DshellWordCompleter(environ._environ.keys())},
                    {'argument': 'value'},
                    env_set]}

}



from unittest import TestCase
from unittest.mock import patch

from flask_jwt_extended import create_access_token
from pkg_resources import parse_version

from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from dm.web.background_tasks import upgrade_version, catalog_logger
from dm.web.network import Response


class Test(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.background_tasks.get_auth_root')
    @patch('dm.web.background_tasks.run_elevator')
    @patch('dm.web.background_tasks.get_software')
    @patch('dm.web.background_tasks.dm_version', '1.0')
    def test_upgrade_version(self, mock_get_software, mock_run_elevator, mock_auth):
        ret = upgrade_version({1: Response(msg=dict(version='0.9'), code=200),
                               2: Response(msg=dict(version='1.0'), code=200)})
        self.assertFalse(ret)

        mock_auth.return_value = 'root'
        mock_get_software.return_value = (None, None)
        ret = upgrade_version({1: Response(msg=dict(version='1.1'), code=200),
                               2: Response(msg=dict(version='1.0'), code=200)})
        self.assertFalse(ret)
        mock_get_software.assert_called_once_with(1, 'root')
        mock_run_elevator.assert_not_called()

        mock_get_software.return_value = ('file', '1.1')
        ret = upgrade_version({1: Response(msg=dict(version='1.1'), code=200),
                               2: Response(msg=dict(version='1.0'), code=200)})
        self.assertTrue(ret)
        mock_run_elevator.assert_called_once_with('file', parse_version('1.1'), catalog_logger)



from unittest import TestCase

from dm.domain.entities import Server, Route
from dm.web import create_app, db


class TestServer(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_route(self):
        dest = Server('dest', port=80)
        proxy = Server('proxy', port=80)

        r = Route(destination=dest)

        self.assertEqual(dest, r.destination)
        self.assertIsNone(r.proxy_server)
        self.assertIsNone(r.gate)
        self.assertIsNone(r.cost)

        # routes defined with a gate
        ## dest
        r = Route(destination=dest, gate=dest.gates[0])
        self.assertEqual(dest, r.destination)
        self.assertIsNone(r.proxy_server)
        self.assertEqual(dest.gates[0], r.gate)
        self.assertEqual(0, r.cost)

        r = Route(destination=dest, gate=dest.gates[0], cost=0)
        self.assertEqual(dest, r.destination)
        self.assertIsNone(r.proxy_server)
        self.assertEqual(dest.gates[0], r.gate)
        self.assertEqual(0, r.cost)

        with self.assertRaises(ValueError):
            r = Route(destination=dest, gate=dest.gates[0], cost=1)

        ## proxy
        with self.assertRaises(ValueError):
            Route(destination=dest, gate=proxy.gates[0])

        with self.assertRaises(ValueError):
            Route(destination=dest, gate=proxy.gates[0], cost=0)

        r = Route(destination=dest, gate=proxy.gates[0], cost=1)
        self.assertEqual(dest, r.destination)
        self.assertIsNone(r.proxy_server)
        self.assertEqual(proxy.gates[0], r.gate)
        self.assertEqual(1, r.cost)

        # routes defined with a proxy server
        ## dest
        with self.assertRaises(ValueError):
            r = Route(destination=dest, proxy_server=dest)

        with self.assertRaises(ValueError):
            r = Route(destination=dest, proxy_server=dest, cost=0)

        with self.assertRaises(ValueError):
            r = Route(destination=dest, proxy_server=dest, cost=1)

        ## proxy
        with self.assertRaises(ValueError):
            r = Route(destination=dest, proxy_server=proxy)

        with self.assertRaises(ValueError):
            r = Route(destination=dest, proxy_server=proxy, cost=0)

        r = Route(destination=dest, proxy_server=proxy, cost=1)
        self.assertEqual(dest, r.destination)
        self.assertEqual(proxy, r.proxy_server)
        self.assertIsNone(r.gate)
        self.assertEqual(1, r.cost)


    def test_to_json_proxy_remote(self):
        dest = Server('dest', port=8000)
        proxy = Server('proxy', port=8000)
        r = Route(destination=dest, proxy_server=proxy, cost=1)

        db.session.add_all([dest, proxy])
        db.session.commit()

        self.assertDictEqual({'destination_id': str(dest.id), 'gate_id': None,
                              'proxy_server_id': str(proxy.id), 'cost': 1}, r.to_json())



from flask import request
from flask_jwt_extended import jwt_required
from flask_restful import Resource

from dm.domain.entities import Server
from dm.web import errors, db
from dm.web.decorators import securizer, forward_or_dispatch, lock_catalog, validate_schema
from dm.web.helpers import filter_query, check_param_in_uri
from dm.web.json_schemas import server_patch


class ServerList(Resource):

    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self):
        query = filter_query(Server, request.args)
        return [at.to_json(add_gates=check_param_in_uri('gates'), human=check_param_in_uri('human')) for at in
                query.all()]


class ServerResource(Resource):
    @forward_or_dispatch
    @jwt_required
    @securizer
    def get(self, server_id):
        return Server.query.get_or_404(server_id).to_json(add_gates=check_param_in_uri('gates'),
                                                          human=check_param_in_uri('human'))

    @forward_or_dispatch
    @jwt_required
    @securizer
    @validate_schema(server_patch)
    @lock_catalog
    def patch(self, server_id):
        json_data = request.get_json()

        server = Server.query.get_or_404(server_id)
        new_granules = json_data.get('granules', [])
        if 'all' in new_granules:
            raise errors.KeywordReserved("'all' is a reserved granule")

        server.granules = list(set(server.granules) | set(new_granules))

        for gate in json_data.get('gates'):
            server.add_new_gate(gate['dns_or_ip'], gate['port'], gate.get('hidden'))

        db.session.commit()

        return {}, 204




from unittest import TestCase

from dm.utils.dag import DAG, DAGError


class TestDAG(TestCase):
    def test_nodes(self):
        G = DAG()
        self.assertListEqual(G.nodes, [])

        G.add_node(1)
        self.assertListEqual(G.nodes, [1])

    def test_pred(self):
        G = DAG()
        self.assertDictEqual(G.pred, {})

        G.add_node(1)
        self.assertDictEqual(G.pred, {1: []})

    def test_succ(self):
        G = DAG()
        self.assertDictEqual(G.succ, {})

        G.add_node(1)
        self.assertDictEqual(G.succ, {1: []})

    def test_add_node(self):
        g = DAG()
        g.add_node(1)
        self.assertListEqual(g.nodes, [1])

        class Foo():
            pass

        bar = Foo()
        g.add_node(bar)

        self.assertListEqual(g.nodes, [1, bar])

        g.add_node(bar)

        self.assertListEqual(g.nodes, [1, bar])

    def test_add_nodes_from(self):
        g = DAG()
        g.add_nodes_from(range(3))
        self.assertListEqual(g.nodes, [0, 1, 2])

    def test_add_edge(self):
        class Foo():
            pass

        g = DAG()
        g.add_edge(2, 3)
        bar = Foo()
        g.add_edge(3, bar)
        g.add_edge(3, bar)
        self.assertListEqual(g.nodes, [2, 3, bar])

        self.assertDictEqual(g.succ, {2: [3], 3: [bar], bar: []})

    def test_add_edges_from(self):
        class Foo():
            pass

        bar = Foo()
        g = DAG()
        g.add_edges_from([(3, bar), (2, 3), (2, 3)])

        self.assertDictEqual(g.succ, {2: [3], 3: [bar], bar: []})

    def test_remove_node(self):
        class Foo():
            pass

        bar = Foo()
        g = DAG()
        g.add_edges_from([(3, bar), (2, 3), (2, 3)])

        with self.assertRaises(DAGError):
            g.remove_node(5)

        g.remove_node(3)

        self.assertListEqual(g.nodes, [bar, 2])

        self.assertDictEqual(g.succ, {2: [], bar: []})
        self.assertDictEqual(g.pred, {2: [], bar: []})

    def test_remove_nodes_from(self):
        class Foo():
            pass

        bar = Foo()
        g = DAG()
        g.add_edges_from([(3, bar), (2, 3), (2, 3)])

        g.remove_nodes_from([5, 2, bar])

        self.assertListEqual(g.nodes, [3])

        self.assertDictEqual(g.succ, {3: []})
        self.assertDictEqual(g.pred, {3: []})

    def test_remove_edge(self):
        class Foo():
            pass

        bar = Foo()
        g = DAG()
        g.add_edges_from([(3, bar), (2, 3), (2, 3)])

        g.remove_edge(2, 3)

        self.assertListEqual(g.succ[2], [])

    def test_remove_edges_from(self):
        class Foo():
            pass

        bar = Foo()
        g = DAG()
        g.add_edges_from([(3, bar), (2, 3), (2, 3)])
        ebunch = [(1, 2), (2, 3)]
        g.remove_edges_from(ebunch)

        self.assertListEqual(g.succ[2], [])

    def test_level(self):
        g = DAG([(1, 2), (1, 7), (7, 2), (2, 4), (2, 3), (4, 5), (4, 6)])

        self.assertEqual(g.level(1), 1)
        self.assertEqual(g.level(2), 3)
        self.assertEqual(g.level(3), 4)
        self.assertEqual(g.level(4), 4)
        self.assertEqual(g.level(5), 5)
        self.assertEqual(g.level(6), 5)
        self.assertEqual(g.level(7), 2)

    def test_get_nodes_at_level(self):
        g = DAG([(1, 2), (1, 7), (7, 2), (2, 4), (2, 3), (4, 5), (4, 6)])

        self.assertListEqual([4, 3], g.get_nodes_at_level(4))

    def test_depth(self):
        g = DAG([(1, 2), (1, 7), (7, 2), (2, 4), (2, 3), (4, 5), (4, 6)])

        self.assertEqual(5, g.depth)

    def test_is_cyclic(self):
        g = DAG([(2, 0), (2, 3), (0, 1), (0, 2)])
        self.assertTrue(g.is_cyclic())

        g = DAG([(2, 0), (2, 3), (0, 1), (1, 2)])
        self.assertTrue(g.is_cyclic())

        g = DAG([(2, 0), (2, 3), (0, 1), (3, 3)])
        self.assertTrue(g.is_cyclic())

        g = DAG([(2, 0), (2, 3), (0, 1), (4, 5), (5, 4)])
        self.assertTrue(g.is_cyclic())

    def test_to_dict_of_lists(self):
        g = DAG([(1, 2), (2, 3), (2, 4)])
        self.assertDictEqual(g.to_dict_of_lists(), {1: [2], 2: [3, 4], 3: [], 4: []})

    def test_from_dict_of_lists(self):
        g = DAG.from_dict_of_lists({1: [2], 2: [3, 4], 3: [], 4: [], 5: []})
        self.assertDictEqual(g.pred, {1: [], 2: [1], 3: [2], 4: [2], 5: []})

    def test_root(self):
        g = DAG.from_dict_of_lists({1: [2], 2: [3, 4], 3: [], 4: [], 5: []})
        self.assertListEqual(g.root, [1, 5])

    def test_copy(self):
        g1 = DAG.from_dict_of_lists({1: [2], 2: [3, 4], 3: [], 4: [], 5: []})
        g2 = g1.copy()
        g2.add_edges_from([(2, 5), (2, 6)])

        self.assertDictEqual(g1.succ, {1: [2], 2: [3, 4], 3: [], 4: [], 5: []})
        self.assertDictEqual(g1.pred, {1: [], 2: [1], 3: [2], 4: [2], 5: []})
        self.assertListEqual(g1.nodes, [1, 2, 3, 4, 5])

        self.assertDictEqual(g2.succ, {1: [2], 2: [3, 4, 5, 6], 3: [], 4: [], 5: [], 6: []})
        self.assertDictEqual(g2.pred, {1: [], 2: [1], 3: [2], 4: [2], 5: [2], 6: [2]})
        self.assertListEqual(g2.nodes, [1, 2, 3, 4, 5, 6])

    def test_subtree(self):
        g = DAG([(1, 2), (1, 7), (7, 2), (2, 4), (2, 3), (4, 5), (4, 6)])

        self.assertDictEqual(g.subtree([2]), {2: [4, 3], 3: [], 4: [5, 6], 5: [], 6: []})
        self.assertDictEqual(g.subtree([2, 7]), {7: [2], 2: [4, 3], 3: [], 4: [5, 6], 5: [], 6: []})
        self.assertDictEqual(g.subtree([5, 6]), {5: [], 6: []})



import typing as t

from flask_jwt_extended import create_access_token

from dm import defaults
from dm.domain.entities import Scope, Server, User
from dm.network.auth import HTTPBearerAuth


def get_servers_from_scope(scope: Scope, *args, **kwargs) -> t.List[Server]:
    """
    Returns the servers to lock for the related scope

    Parameters
    ----------
    scope: Scope

    Returns
    -------

    """
    quorum = []
    if scope == scope.CATALOG:
        servers = Server.query.order_by(Server.last_modified_at).all()
        if len(servers) < defaults.MIN_SERVERS_QUORUM:
            return servers
        else:
            cost_dict = {}
            for server in servers:
                if server.route and server.route.cost is not None:
                    if server.route.cost not in cost_dict:
                        cost_dict[server.route.cost] = []
                    cost_dict[server.route.cost].append(server)
            if len(cost_dict) > defaults.MIN_SERVERS_QUORUM:
                for v in cost_dict.values():
                    v.sort(key=lambda x: (x.last_modified_at, x.name))
                    quorum.append(v[0])
            else:
                quorum.extend(servers[0:defaults.MIN_SERVERS_QUORUM])
            me = Server.get_current()
            if me not in quorum:
                quorum.append(me)
    elif scope == scope.UPGRADE:
        quorum = Server.get_current()
    return quorum


def get_auth_root():
    return HTTPBearerAuth(create_access_token(User.get_by_user('root').id))



import os
import sys

from prompt_toolkit.formatted_text import is_formatted_text


def get_raw_text(text):
    if is_formatted_text(text):
        return ''.join([token[1] for token in text.formatted_text])
    else:
        text


def clean_none(data):
    return {k: v for k, v in data.items() if v is not None}


def is_interactive(output=False, error=False, heuristic=False):
    if not sys.stdin.isatty():
        return False
    if output and not sys.stdout.isatty():
        return False
    if error and not sys.stderr.isatty():
        return False

    if heuristic:
        home = os.getenv('HOME')
        homepath = os.getenv('HOMEPATH')
        if not homepath and (not home or home == '/'):
            return False
    return True



import datetime as dt
import ipaddress
import json
import pickle
import typing as t
import uuid

import rsa
from dateutil.tz import tzlocal
from sqlalchemy import types, DateTime, Enum as _Enum
from sqlalchemy.dialects.postgresql import UUID as pUUID

Kwargs = t.Union[t.Mapping['str', t.Any]]
Id = t.TypeVar('Id', int, str, uuid.UUID, tuple)
Ids = t.Sequence[Id]
Id_or_Ids = t.Union[Id, Ids]

IPs = t.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

Gate = t.Union[t.Tuple[t.Union[str, IPs], int]]

tJSON = t.Union[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]

Callback = t.Tuple[t.Callable[[], None], t.Tuple, t.Dict]

Priority = t.TypeVar('T')


class TypeDecorator(types.TypeDecorator):
    def __repr__(self):
        return self.impl.__repr__()


class ScalarListException(Exception):
    pass


class ScalarListType(TypeDecorator):
    """
    ScalarListType type provides convenient way for saving multiple scalar
    values in one column. ScalarListType works like list on python side and
    saves the result as comma-separated list in the database (custom separators
    can also be used).

    Example ::


        from sqlalchemy_utils import ScalarListType


        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, autoincrement=True)
            hobbies = sa.Column(ScalarListType())


        user = User()
        user.hobbies = [u'football', u'ice_hockey']
        session.commit()


    You can easily set up integer lists too:

    ::


        from sqlalchemy_utils import ScalarListType


        class Player(Base):
            __tablename__ = 'player'
            id = sa.Column(sa.Integer, autoincrement=True)
            points = sa.Column(ScalarListType(int))


        player = Player()
        player.points = [11, 12, 8, 80]
        session.commit()


    """

    impl = types.Text()

    def __init__(self, coerce_func=str, separator=':'):
        self.separator = separator
        self.coerce_func = coerce_func

    def process_bind_param(self, value, dialect):
        # Convert list of values to unicode separator-separated list
        # Example: [1, 2, 3, 4] -> u'1, 2, 3, 4'
        if value is not None:
            if any(self.separator in str(item) for item in value):
                raise ScalarListException(
                    "List values can't contain string '%s' (its being used as "
                    "separator. If you wish for scalar list values to contain "
                    "these strings, use a different separator string.)"
                    % self.separator
                )
            return self.separator.join(
                map(str, value)
            )

    def process_result_value(self, value, dialect):
        if value is not None:
            if value == u'':
                return []
            # coerce each value
            return list(map(
                self.coerce_func, value.split(self.separator)
            ))


class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(36)

    """
    impl = types.CHAR(36)

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(pUUID())
        else:
            return dialect.type_descriptor(types.CHAR(36))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        return str(value).lower()

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        return value.lower()


class Enum(_Enum):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """
    impl = types.VARCHAR(80)



class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class PrivateKey(TypeDecorator):
    impl = types.BLOB

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = value.save_pkcs1()
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = rsa.PrivateKey.load_pkcs1(value)
        return value


class PublicKey(TypeDecorator):
    impl = types.BLOB

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = value.save_pkcs1()
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = rsa.PublicKey.load_pkcs1(value)
        return value


class IP(TypeDecorator):
    impl = types.VARCHAR(39)

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = str(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = ipaddress.ip_address(value)
        return value


class Pickle(TypeDecorator):
    impl = types.BLOB

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = pickle.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = pickle.loads(value)
        return value


class Utc(dt.tzinfo):
    __slots__ = ()

    zero = dt.timedelta(0)

    def utcoffset(self, _):
        return self.zero

    def dst(self, _):
        return self.zero

    def tzname(self, _):
        return 'UTC'


try:
    utc = dt.timezone.utc
except AttributeError:
    utc = Utc()


class UtcDateTime(TypeDecorator):
    """Almost equivalent to :class:`~sqlalchemy.types.DateTime` with
    ``timezone=True`` option, but it differs from that by:
    - Never silently take naive :class:`~datetime.datetime`, instead it
      always raise :exc:`ValueError` unless time zone aware value.
    - :class:`~datetime.datetime` value's :attr:`~datetime.datetime.tzinfo`
      is always converted to UTC.
    - Unlike SQLAlchemy's built-in :class:`~sqlalchemy.types.DateTime`,
      it never return naive :class:`~datetime.datetime`, but time zone
      aware value, even with SQLite or MySQL.
    """

    impl = DateTime(timezone=True)

    def process_bind_param(self, value, dialect):
        if value is not None:
            if not isinstance(value, dt.datetime):
                raise TypeError('expected datetime.datetime, not ' +
                                repr(value))
            elif value.tzinfo is None:
                raise ValueError('naive datetime is disallowed')
            return value.astimezone(utc)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        elif value.tzinfo is None:
            value = value.replace(tzinfo=utc)
        return value.astimezone(tzlocal())


import abc
import inspect
import random
import string
import uuid
from datetime import datetime

from sqlalchemy import Column, Boolean

from dm import defaults
from dm.utils.typos import UUID, UtcDateTime


class JSONEntity:

    @abc.abstractmethod
    def to_json(self):
        ...

    @classmethod
    @abc.abstractmethod
    def from_json(cls, kwargs):
        ...


class SoftDeleteMixin:
    __prefix__ = '_old_'
    deleted = Column(Boolean(), default=False)

    def __init__(self, deleted=False, **kwargs):
        self.deleted = deleted
        for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
            setattr(self, attr, kwargs.get(attr, None))

    def delete(self):
        if not self.deleted:
            self.deleted = True
            for attr in [attr for attr, value in inspect.getmembers(self) if attr.startswith(self.__prefix__)]:
                original_attr = attr.lstrip(self.__prefix__)
                setattr(self, attr, getattr(self, original_attr))
                setattr(self, original_attr,
                        ''.join(random.choices(string.digits + string.ascii_letters + string.punctuation, k=10)))



class DistributedEntityMixin(JSONEntity):
    order = None
    last_modified_at = Column(UtcDateTime(), nullable=False)

    def __init__(self, **kwargs):
        self.last_modified_at = kwargs.pop('last_modified_at', None)

    def to_json(self):
        try:
            return dict(last_modified_at=self.last_modified_at.strftime(defaults.DATEMARK_FORMAT))
        except AttributeError:
            return dict()

    @classmethod
    def from_json(cls, kwargs):
        if 'last_modified_at' in kwargs:
            last_modified_at = kwargs.pop('last_modified_at')
            last_modified_at = datetime.strptime(last_modified_at, defaults.DATEMARK_FORMAT)
            kwargs.update(last_modified_at=last_modified_at)


class UUIDEntityMixin:
    id = Column(UUID, primary_key=True, default=str(uuid.uuid4))

    def __init__(self, **kwargs):
        if 'id' in kwargs:
            self.id = str(kwargs['id']).lower()
        else:
            self.id = str(uuid.uuid4())


class UUIDistributedEntityMixin(UUIDEntityMixin, DistributedEntityMixin):

    def __init__(self, **kwargs):
        UUIDEntityMixin.__init__(self, **kwargs)
        DistributedEntityMixin.__init__(self, **kwargs)

    def to_json(self):
        data = super().to_json()
        if self.id:
            data['id'] = str(self.id)
        return data

    @classmethod
    @abc.abstractmethod
    def from_json(cls, kwargs):
        super().from_json(kwargs)
        try:
            o = cls.query.get(kwargs.get('id'))
        except RuntimeError as e:
            o = None
        if o:
            for k, v in kwargs.items():
                if getattr(o, k) != v:
                    setattr(o, k, v)
            return o
        else:
            return cls(**kwargs)



class EntityReprMixin:
    id = None

    def __repr__(self):
        if self.id:
            return f'<{self.__class__.__name__} {self.id}>'
        else:
            return f'<{self.__class__.__name__} (transient {id(self)})>'

    def __str__(self):
        return self.__repr__()



import typing as t
from datetime import datetime

from flask_jwt_extended import get_jwt_identity
from passlib.hash import sha256_crypt

from dm import defaults
from dm.domain.entities.base import UUIDistributedEntityMixin
from dm.utils.helpers import get_now
from dm.utils.typos import ScalarListType, UtcDateTime
from dm.web import db


class User(db.Model, UUIDistributedEntityMixin):
    __tablename__ = 'D_user'

    user = db.Column(db.String(30), nullable=False)
    _password = db.Column('password', db.String(256))
    email = db.Column(db.String)
    created_at = db.Column(UtcDateTime)
    active = db.Column('is_active', db.Boolean(), nullable=False)
    groups = db.Column(ScalarListType(str))

    __table_args__ = (db.UniqueConstraint('user', name='D_user_uq01'),)

    def __init__(self, user, password=None, email=None, created_at=None, active=True, groups: t.Union[str, t.List[str]] = None,
                 **kwargs):

        UUIDistributedEntityMixin.__init__(self, **kwargs)
        self.user = user
        self._password = password or kwargs.get('_password', None)
        self.email = email
        self.created_at = created_at or get_now()
        self.active = active
        if isinstance(groups, str):
            self.groups = groups.split(':')
        else:
            self.groups = groups or []

    @classmethod
    def get_by_user(cls, user):
        return db.session.query(cls).filter_by(user=user).one_or_none()

    @classmethod
    def get_by_group(cls, group):
        return [g for g in db.session.query(cls).all() if group in g.groups]

    def _hash_password(self, password):
        if not self._password:
            self._password = sha256_crypt.hash(password)

    def verify_password(self, password) -> bool:
        return sha256_crypt.verify(password, self._password)

    def set_password(self, password):
        self._password = None
        self._hash_password(password)

    def to_json(self, password=False) -> dict:
        data = super().to_json()
        data.update(user=self.user, email=self.email, created_at=self.created_at.strftime(defaults.DATETIME_FORMAT),
                    active=self.active, groups=','.join(self.groups))
        if password:
            data.update(_password=self._password)
        return data

    @classmethod
    def from_json(cls, kwargs):
        kwargs = dict(kwargs)
        if 'created_at' in kwargs:
            kwargs['created_at'] = datetime.strptime(kwargs.get('created_at'), defaults.DATETIME_FORMAT)
        if 'groups' in kwargs:
            kwargs['groups'] = kwargs['groups'].split(':')
        return super().from_json(kwargs)

    @classmethod
    def set_initial(cls):
        from dm.domain.entities import bypass_datamark_update
        with bypass_datamark_update():
            root = cls.get_by_user('root')
            if not root:
                root = User(id='00000000-0000-0000-0000-000000000001', user='root', groups=['administrator'],
                            last_modified_at=defaults.INITIAL_DATEMARK)
                root.set_password('')
                db.session.add(root)
            ops = cls.get_by_user('ops')
            if not ops:
                ops = User(id='00000000-0000-0000-0000-000000000002', user='ops', groups=['operator', 'deployer'],
                           last_modified_at=defaults.INITIAL_DATEMARK)
                db.session.add(ops)
            reporter = cls.get_by_user('reporter')
            if not reporter:
                reporter = User(id='00000000-0000-0000-0000-000000000003', user='reporter', groups=['readonly'],
                                last_modified_at=defaults.INITIAL_DATEMARK)
                db.session.add(reporter)
            join = cls.get_by_user('join')
            if not join:
                join = User(id='00000000-0000-0000-0000-000000000004', user='join', groups=[''],
                                last_modified_at=defaults.INITIAL_DATEMARK)
                db.session.add(join)

    @classmethod
    @staticmethod
    def get_current():
        return db.session.query(User).get(get_jwt_identity())

    def __repr__(self):
        return f"{self.id}.{self.user}"

    def __str__(self):
        return self.user



import os
from unittest import mock
from unittest.mock import patch

import requests
import responses
from pkg_resources import parse_version
from pyfakefs.fake_filesystem_unittest import TestCase

from dm.domain.entities import Dimension
from dm.domain.entities.bootstrap import set_initial
from dm.web import create_app, db
from dm.web.background_tasks import process_get_new_version_from_gogs, \
    upgrader_logger

gogs_content = """
<div class="ui container">
		<h2 class="ui header">
			Releases
		</h2>
		<ul id="release-list">
				<li class="ui grid">
					<div class="ui four wide column meta">
						<span class="commit">
							<a href="/dimensigon/dimensigon/src/d3aad4973fe692c4fcefede8eb53a1c9e32749b2" rel="nofollow"><i class="code icon"></i> d3aad4973f</a>
						</span>
					</div>
					<div class="ui twelve wide column detail">
							<h4>
								<a href="/dimensigon/dimensigon/src/v0.1.a1" rel="nofollow"><i class="tag icon"></i> v0.1.a1</a>
							</h4>
							<div class="download">
								<a href="/dimensigon/dimensigon/archive/v0.1.a1.zip" rel="nofollow"><i class="octicon octicon-file-zip"></i>ZIP</a>
								<a href="/dimensigon/dimensigon/archive/v0.1.a1.tar.gz"><i class="octicon octicon-file-zip"></i>TAR.GZ</a>
							</div>
						<span class="dot">&nbsp;</span>
					</div>
				</li>
				<li class="ui grid">
					<div class="ui four wide column meta">
						<span class="commit">
							<a href="/dimensigon/dimensigon/src/2184389034cec9620c44594ae6c174e676434db5" rel="nofollow"><i class="code icon"></i> 2184389034</a>
						</span>
					</div>
					<div class="ui twelve wide column detail">
							<h4>
								<a href="/dimensigon/dimensigon/src/v0.0.1" rel="nofollow"><i class="tag icon"></i> v0.0.1</a>
							</h4>
							<div class="download">
								<a href="/dimensigon/dimensigon/archive/v0.0.1.zip" rel="nofollow"><i class="octicon octicon-file-zip"></i>ZIP</a>
								<a href="/dimensigon/dimensigon/archive/v0.0.1.tar.gz"><i class="octicon octicon-file-zip"></i>TAR.GZ</a>
							</div>
						<span class="dot">&nbsp;</span>
					</div>
				</li>
		</ul>
		<div class="center">
			<a class="ui small button disabled">
				Página Anterior
			</a>
			<a class="ui small button disabled">
				Página Siguiente
			</a>
		</div>
	</div>
"""


class TestProcessGetNewVersionFromGogs(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        with self.app.app_context():
            db.create_all()
            set_initial()
            d = Dimension(name='test', current=True)
            db.session.add(d)
            db.session.commit()
        self.client = self.app.test_client(use_cookies=True)
        self.setUpPyfakefs()
        self.fs.create_dir(self.app.config['SOFTWARE_REPO'])

    def tearDown(self) -> None:
        with self.app.app_context():
            db.session.remove()
            db.drop_all()

    @patch('dm.web.background_tasks.run_elevator')
    @responses.activate
    def test_internet_upgrade(self, mock_run_elevator):
        with mock.patch('dm.web.background_tasks.dm_version', '0.0.1'):
            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/releases',
                          body=gogs_content)

            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/archive/v0.1.a1.tar.gz',
                          body=b"v0.1.a1")

            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/archive/v0.0.1.tar.gz',
                          body=b"v0.0.1")

            process_get_new_version_from_gogs(self.app)

            self.assertTrue(os.path.exists(
                os.path.join(self.app.config['SOFTWARE_REPO'], 'dimensigon', 'dimensigon-v0.1.a1.tar.gz')))
            self.assertEqual(
                (os.path.join(self.app.config['SOFTWARE_REPO'], 'dimensigon', 'dimensigon-v0.1.a1.tar.gz'),
                 parse_version('v0.1.a1'),
                 upgrader_logger),
                mock_run_elevator.call_args[0])

    @patch('dm.web.background_tasks.run_elevator')
    @responses.activate
    def test_internet_not_upgrade(self, mock_run_elevator):
        with mock.patch('dm.web.background_tasks.dm_version', '0.1'):
            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/releases',
                          body=gogs_content)

            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/archive/v0.1.a1.tar.gz',
                          body=b"v0.1.a1")

            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/archive/v0.0.1.tar.gz',
                          body=b"v0.0.1")

            process_get_new_version_from_gogs(self.app)

            self.assertFalse(mock_run_elevator.called)

    @patch('dm.web.background_tasks.run_elevator')
    @responses.activate
    def test_no_internet_no_upgrade(self, mock_run_elevator):
        with mock.patch('dm.web.background_tasks.dm_version', '0.1'):
            responses.add(method='GET',
                          url=self.app.config['GIT_REPO'] + '/dimensigon/dimensigon/releases',
                          body=requests.exceptions.ConnectionError('No connection'))

            process_get_new_version_from_gogs(self.app)

            self.assertFalse(mock_run_elevator.called)



from unittest import TestCase

from dm.domain.entities import Server
from dm.web import create_app, db
from dm.web import errors


class TestBaseQuery(TestCase):

    def setUp(self) -> None:
        self.app = create_app('test')

    def test_get_or_404(self):
        with self.app.app_context():
            db.create_all()
            with self.assertRaises(errors.EntityNotFound) as cm:
                Server.query.get_or_404(1)

            self.assertTupleEqual(cm.exception.args, ("Server", 1))

    def test_first_or_404(self):
        with self.app.app_context():
            db.create_all()
            with self.assertRaises(errors.NoDataFound) as cm:
                Server.query.first_or_404()

            self.assertEqual(cm.exception.args, ("Server",))



import os
import time

from asynctest import patch, TestCase
from testfixtures import LogCapture

from dm.domain.entities.log import Log
from dm.utils.helpers import encode
from dm.web import create_app, repo, interactor
from tests.helpers import set_response_from_mock, wait_mock_called
from tests.system.data import Server1, Server2

DEST_FOLDER = os.path.dirname(os.path.abspath(__file__))


class TestSendDataLog(TestCase):

    def setUp(self) -> None:
        self.file1 = os.path.join(DEST_FOLDER, 'server1.tempfile.log')
        self.file2 = os.path.join(DEST_FOLDER, 'server2.tempfile.log')
        self.remove_files()

        self.app1 = create_app(Server1())
        self.app2 = create_app(Server2())

        self.client1 = self.app1.test_client()
        self.client2 = self.app2.test_client()
        self.lines = ['line 1\nline 2\n', 'line 3\n', 'line 4\nline 5\n']
        self.i = 0
        self.append_data()

    def append_data(self):
        with open(self.file1, 'a') as temp:
            temp.write(self.lines[self.i])
        self.i += 1

    def get_file_offset(self):
        try:
            with open(self.file1 + '.offset', 'r') as temp:
                return int(temp.readlines()[1].strip())
        except:
            return None

    def get_current_offset(self):
        with open(self.file1, 'r') as fd:
            fd.seek(0, 2)
            return fd.tell()

    def remove_files(self):
        for file in (self.file1, self.file2):
            try:
                os.remove(file)
            except:
                pass
            try:
                os.remove(file + '.offset')
            except:
                pass

    def tearDown(self, c=0) -> None:
        with self.app1.app_context():
            interactor.stop_send_data_logs()
        self.remove_files()

    @patch('dm.network.encryptation.requests.post')
    def test_send_data_log(self, mock_post):
        set_response_from_mock(mock_post, url='http://server2.localdomain:81/socket?', status=200, json='')
        with self.app1.app_context():
            log = Log(file=self.file1, server=repo.ServerRepo.find('bbbbbbbb-1234-5678-1234-56781234bbb2'),
                      dest_folder=DEST_FOLDER, dest_name=os.path.basename(self.file2))
            repo.LogRepo.add(log)
            del log

            interactor._delay = None
            resp = interactor.send_data_logs(blocking=False, delay=None)
            wait_mock_called(mock_post, 1, 10)

            mock_post.assert_called_once_with('http://server2.localdomain:81/socket',
                                              json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                                    'data': encode(filename=os.path.basename(self.file2),
                                                                   data_log=self.lines[0], dest_folder=DEST_FOLDER)})

        with self.app2.app_context():
            self.client2.post('/socket', json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                               'data': encode(filename=os.path.basename(self.file2),
                                                              data_log=self.lines[0], dest_folder=DEST_FOLDER)})
            c = 0
            while not os.path.exists(self.file2) and c < 50:
                time.sleep(0.01)
                c += 1
            self.assertTrue(os.path.exists(self.file2))
            with open(self.file2) as fd:
                self.assertEqual(self.lines[0], fd.read())

        with self.app1.app_context():
            self.append_data()
            # force to awake thread
            interactor._awake.set()
            # wait until it reads the new data
            wait_mock_called(mock_post, 2, 10)
            interactor._awake.clear()
            mock_post.assert_called_with('http://server2.localdomain:81/socket',
                                         json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                               'data': encode(filename=os.path.basename(self.file2),
                                                              data_log=self.lines[1], dest_folder=DEST_FOLDER)})

        with self.app2.app_context():
            self.client2.post('/socket', json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                               'data': encode(filename=os.path.basename(self.file2),
                                                              data_log=self.lines[1], dest_folder=DEST_FOLDER)})
            c = 0
            while not os.path.exists(self.file2) and c < 50:
                time.sleep(0.01)
                c += 1
            self.assertTrue(os.path.exists(self.file2))
            with open(self.file2) as fd:
                self.assertEqual(''.join(self.lines[0:2]), fd.read())

    @patch('dm.network.encryptation.requests.post')
    def test_send_data_log_with_error(self, mock_post):

        with self.app1.app_context():
            log = Log(file=self.file1, server=repo.ServerRepo.find('bbbbbbbb-1234-5678-1234-56781234bbb2'),
                      dest_folder=DEST_FOLDER, dest_name=os.path.basename(self.file2))
            repo.LogRepo.add(log)
            del log

            set_response_from_mock(mock_post, url='http://server2.localdomain:81/socket?', status=500,
                                   json='{"error": "Permission Denied"}')
            interactor._delay = None

            with LogCapture() as l:
                resp = interactor.send_data_logs(blocking=False, delay=None)
                wait_mock_called(mock_post, 1, 50)
                self.assertTrue(True)

            self.assertFalse(os.path.exists(self.file1 + '.offset'))

            mock_post.assert_called_once_with('http://server2.localdomain:81/socket',
                                              json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                                    'data': encode(filename=os.path.basename(self.file2),
                                                                   data_log=self.lines[0], dest_folder=DEST_FOLDER)})

            set_response_from_mock(mock_post, url='http://server2.localdomain:81/socket?', status=200, json='')

            self.append_data()
            # force to awake thread
            interactor._awake.set()
            # wait until it reads the new data
            wait_mock_called(mock_post, 2, 50)
            interactor._awake.clear()
            mock_post.assert_called_with('http://server2.localdomain:81/socket',
                                         json={'destination': 'bbbbbbbb-1234-5678-1234-56781234bbb2',
                                               'data': encode(filename=os.path.basename(self.file2),
                                                              data_log=''.join(self.lines[0:2]),
                                                              dest_folder=DEST_FOLDER)})



import datetime as dt
from unittest.mock import patch

from flask import url_for

from dm import defaults
from dm.domain.entities import Server, ActionTemplate, ActionType
from dm.web import db
from tests.helpers import TestDimensigonBase


class TestApi(TestDimensigonBase):

    @patch('dm.web.api_1_0.urls.use_cases.get_distributed_entities')
    @patch('dm.domain.entities.get_now')
    def test_catalog(self, mock_now, mock_get):
        mock_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        mock_get.return_value = [('ActionTemplate', ActionTemplate)]
        # add data
        s = Server('test', post=8000, me=True)
        db.session.add(s)

        at1 = ActionTemplate(name='ActionTest1', version=1, action_type=ActionType.ORCHESTRATION, code='')
        db.session.add(at1)
        db.session.commit()

        resp = self.client.get(
            url_for('api_1_0.catalog',
                    data_mark=dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc).strftime(defaults.DATEMARK_FORMAT)),
            headers=self.auth.header)

        self.assertDictEqual({'ActionTemplate': [at1.to_json()]},
                             resp.get_json())

        mock_now.return_value = dt.datetime(2019, 4, 2, 1, tzinfo=dt.timezone.utc)
        at2 = ActionTemplate(name='ActionTest2', version=1, action_type=ActionType.ORCHESTRATION, code='')
        db.session.add(at2)
        db.session.commit()

        resp = self.client.get(
            url_for('api_1_0.catalog',
                    data_mark=dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc).strftime(defaults.DATEMARK_FORMAT)),
            headers=self.auth.header)

        self.assertDictEqual({'ActionTemplate': [at2.to_json()]},
                             resp.get_json())



from unittest import TestCase
from unittest.mock import patch, Mock

from dm.utils.event_handler import EventHandler, _RegistryContainer, Event


class TestEventHandler(TestCase):

    def setUp(self) -> None:
        self.eh = EventHandler()

    @patch('dm.utils.event_handler.time.time')
    def test_discard(self, mock_time):
        mock_time.return_value = 2
        self.eh.discard_after = 2
        self.eh._registry[1] = _RegistryContainer(lambda x: x, (), {}, 1)
        self.eh._registry[2] = _RegistryContainer(lambda x: x, (), {}, 0)
        self.eh._pending_events[1] = (Event(1, data={}), 1)
        self.eh._pending_events[2] = (Event(2, data={}), 0)

        self.eh.discard()

        self.assertEqual(1, len(self.eh._registry))
        self.assertEqual(1, len(self.eh._pending_events))
        self.assertIn(1, self.eh._registry)
        self.assertIn(1, self.eh._pending_events)

    @patch('dm.utils.event_handler.time.time')
    def test_register(self, mock_time):
        func = Mock()
        mock_time.return_value = 0
        self.eh.discard_after = 2
        self.eh.register(1, func)
        self.assertEqual(1, len(self.eh._registry))

        with self.assertRaises(ValueError):
            self.eh.register(1, func)

        e = Event(2, {})
        self.eh._pending_events[2] = (e, 3)
        mock_time.return_value = 3
        func2 = Mock()
        self.eh.register(2, func2, ('arg',), kwargs={'param': 'kwarg'})

        func2.assert_called_once_with(e, 'arg', param='kwarg')
        self.assertEqual(0, len(self.eh._registry))

    @patch('dm.utils.event_handler.time.time')
    def test_dispatch(self, mock_time):
        func = Mock()
        mock_time.return_value = 0
        self.eh.discard_after = 2
        self.eh.register(1, func)
        e = Event(1, {})
        self.eh.dispatch(e)
        func.assert_called_once_with(e)

        e = Event(2, {})
        self.eh.dispatch(e)
        self.assertIn(2, self.eh._pending_events)

        mock_time.return_value = 2
        e = Event(3, {})
        self.eh.dispatch(e)
        self.assertNotIn(2, self.eh._pending_events)
        self.assertIn(3, self.eh._pending_events)


import concurrent.futures
import re

from flask import copy_current_request_context, has_request_context
from flask.globals import _app_ctx_stack

from dm.web.extensions.flask_executor.futures import FutureCollection, FutureProxy
from dm.web.extensions.flask_executor.helpers import InstanceProxy, str2bool


def copy_current_app_context(fn):
    app_context = _app_ctx_stack.top

    def wrapper(*args, **kwargs):
        with app_context:
            return fn(*args, **kwargs)

    return wrapper


def propagate_exceptions_callback(future):
    exc = future.exception()
    if exc:
        raise exc


class ExecutorJob:
    """Wraps a function with an executor so to allow the wrapped function to
    submit itself directly to the executor."""

    def __init__(self, executor, fn):
        self.executor = executor
        self.fn = fn

    def submit(self, *args, **kwargs):
        future = self.executor.submit(self.fn, *args, **kwargs)
        return future

    def submit_stored(self, future_key, *args, **kwargs):
        future = self.executor.submit_stored(future_key, self.fn, *args, **kwargs)
        return future

    def map(self, *iterables, **kwargs):
        results = self.executor.map(self.fn, *iterables, **kwargs)
        return results


class Executor(InstanceProxy, concurrent.futures._base.Executor):
    """An executor interface for :py:mod:`concurrent.futures` designed for
    working with Flask applications.
    :param app: A Flask application instance.
    :param name: An optional name for the executor. This can be used to
                 configure multiple executors. Named executors will look for
                 environment variables prefixed with the name in uppercase,
                 e.g. ``CUSTOM_EXECUTOR_TYPE``.
    """

    def __init__(self, app=None, name=''):
        self.app = app
        self._default_done_callbacks = []
        self.futures = FutureCollection()
        if re.match(r'^(\w+)?$', name) is None:
            raise ValueError(
                "Executor names may only contain letters, numbers or underscores"
            )
        self.name = name
        prefix = name.upper() + '_' if name else ''
        self.EXECUTOR_TYPE = prefix + 'EXECUTOR_TYPE'
        self.EXECUTOR_MAX_WORKERS = prefix + 'EXECUTOR_MAX_WORKERS'
        self.EXECUTOR_FUTURES_MAX_LENGTH = prefix + 'EXECUTOR_FUTURES_MAX_LENGTH'
        self.EXECUTOR_PROPAGATE_EXCEPTIONS = prefix + 'EXECUTOR_PROPAGATE_EXCEPTIONS'
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Initialise application. This will also intialise the configured
        executor type:
            * :class:`concurrent.futures.ThreadPoolExecutor`
            * :class:`concurrent.futures.ProcessPoolExecutor`
        """
        app.config.setdefault(self.EXECUTOR_TYPE, 'thread')
        futures_max_length = app.config.setdefault(self.EXECUTOR_FUTURES_MAX_LENGTH, None)
        propagate_exceptions = app.config.setdefault(self.EXECUTOR_PROPAGATE_EXCEPTIONS, False)
        if futures_max_length is not None:
            self.futures.max_length = int(futures_max_length)
        if str2bool(propagate_exceptions):
            self.add_default_done_callback(propagate_exceptions_callback)
        self._self = self._make_executor(app)
        app.extensions[self.name + 'executor'] = self

    def _make_executor(self, app):
        executor_max_workers = app.config.setdefault(self.EXECUTOR_MAX_WORKERS, None)
        if executor_max_workers is not None:
            executor_max_workers = int(executor_max_workers)
        executor_type = app.config[self.EXECUTOR_TYPE]
        if executor_type == 'thread':
            _executor = concurrent.futures.ThreadPoolExecutor
        elif executor_type == 'process':
            _executor = concurrent.futures.ProcessPoolExecutor
        else:
            raise ValueError("{} is not a valid executor type.".format(executor_type))
        return _executor(max_workers=executor_max_workers)

    def _prepare_fn(self, fn, force_copy=False):
        if isinstance(self._self, concurrent.futures.ThreadPoolExecutor) or force_copy:
            if has_request_context():
                fn = copy_current_request_context(fn)
            fn = copy_current_app_context(fn)
        return fn

    def submit(self, fn, *args, **kwargs):
        """Schedules the callable, fn, to be executed as fn(\*args \**kwargs)
        and returns a :class:`~flask_executor.futures.FutureProxy` object, a
        :class:`~concurrent.futures.Future` subclass representing
        the execution of the callable.
        See also :meth:`concurrent.futures.Executor.submit`.
        Callables are wrapped a copy of the current application context and the
        current request context. Code that depends on information or
        configuration stored in :data:`flask.current_app`,
        :data:`flask.request` or :data:`flask.g` can be run without
        modification.
        Note: Because callables only have access to *copies* of the application
        or request contexts any changes made to these copies will not be
        reflected in the original view. Further, changes in the original app or
        request context that occur after the callable is submitted will not be
        available to the callable.
        Example::
            future = executor.submit(pow, 323, 1235)
            print(future.result())
        :param fn: The callable to be executed.
        :param \*args: A list of positional parameters used with
                       the callable.
        :param \**kwargs: A dict of named parameters used with
                          the callable.
        :rtype: flask_executor.FutureProxy
        """
        fn = self._prepare_fn(fn)
        future = self._self.submit(fn, *args, **kwargs)
        for callback in self._default_done_callbacks:
            future.add_done_callback(callback)
        return FutureProxy(future, self)

    def submit_stored(self, future_key, fn, *args, **kwargs):
        """Submits the callable using :meth:`Executor.submit` and stores the
        Future in the executor via a
        :class:`~flask_executor.futures.FutureCollection` object available at
        :data:`Executor.futures`. These futures can be retrieved anywhere
        inside your application and queried for status or popped from the
        collection. Due to memory concerns, the maximum length of the
        FutureCollection is limited, and the oldest Futures will be dropped
        when the limit is exceeded.
        See :class:`flask_executor.futures.FutureCollection` for more
        information on how to query futures in a collection.
        Example::
            @app.route('/start-task')
            def start_task():
                executor.submit_stored('calc_power', pow, 323, 1235)
                return jsonify({'result':'success'})
            @app.route('/get-result')
            def get_result():
                if not executor.futures.done('calc_power'):
                    future_status = executor.futures._state('calc_power')
                    return jsonify({'status': future_status})
                future = executor.futures.pop('calc_power')
                return jsonify({'status': done, 'result': future.result()})
        :param future_key: Stores the Future for the submitted task inside the
                           executor's ``futures`` object with the specified
                           key.
        :param fn: The callable to be executed.
        :param \*args: A list of positional parameters used with
                       the callable.
        :param \**kwargs: A dict of named parameters used with
                          the callable.
        :rtype: concurrent.futures.Future
        """
        future = self.submit(fn, *args, **kwargs)
        self.futures.add(future_key, future)
        return future

    def map(self, fn, *iterables, **kwargs):
        """Submits the callable, fn, and an iterable of arguments to the
        executor and returns the results inside a generator.
        See also :meth:`concurrent.futures.Executor.map`.
        Callables are wrapped a copy of the current application context and the
        current request context. Code that depends on information or
        configuration stored in :data:`flask.current_app`,
        :data:`flask.request` or :data:`flask.g` can be run without
        modification.
        Note: Because callables only have access to *copies* of the application
        or request contexts
        any changes made to these copies will not be reflected in the original
        view. Further, changes in the original app or request context that
        occur after the callable is submitted will not be available to the
        callable.
        :param fn: The callable to be executed.
        :param \*iterables: An iterable of arguments the callable will apply to.
        :param \**kwargs: A dict of named parameters to pass to the underlying
                          executor's :meth:`~concurrent.futures.Executor.map`
                          method.
        """
        fn = self._prepare_fn(fn)
        return self._self.map(fn, *iterables, **kwargs)

    def job(self, fn):
        """Decorator. Use this to transform functions into `ExecutorJob`
        instances that can submit themselves directly to the executor.
        Example::
            @executor.job
            def fib(n):
                if n <= 2:
                    return 1
                else:
                    return fib(n-1) + fib(n-2)
            future = fib.submit(5)
            results = fib.map(range(1, 6))
        """
        if isinstance(self._self, concurrent.futures.ProcessPoolExecutor):
            raise TypeError(
                "Can't decorate {}: Executors that use multiprocessing "
                "don't support decorators".format(fn)
            )
        return ExecutorJob(executor=self, fn=fn)

    def add_default_done_callback(self, fn):
        """Registers callable to be attached to all newly created futures. When a
        callable is submitted to the executor,
        :meth:`concurrent.futures.Future.add_done_callback` is called for every default
        callable that has been set."
        :param fn: The callable to be added to the list of default done callbacks for new
                   Futures.
        """

        self._default_done_callbacks.append(fn)


import datetime as dt
import os
from unittest import TestCase
from unittest.mock import patch

import responses
from aioresponses import aioresponses
from flask import url_for
from flask_jwt_extended import create_access_token, verify_jwt_in_request

from dimensigon import Software, ActionTemplate, ActionType, SoftwareServerAssociation, Route, User, Locker
from dm import defaults
from dm.domain.entities import Server, Dimension, Catalog
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db
from dm.web.background_tasks import update_catalog
from dm.web.network import Response
from tests.helpers import set_callbacks

basedir = os.path.abspath(os.path.dirname(__file__))


@patch('dm.web.background_tasks.dm_version', '1')
@patch('dm.web.routes.dm.__version__', '1')
class TestUpdateCatalog(TestCase):

    # @staticmethod
    # def remove_db_files():
    #     with contextlib.suppress(FileNotFoundError):
    #         os.remove(os.path.join(basedir, 'node1.db'))
    #         os.remove(os.path.join(basedir, 'node2.db'))
    #
    # @classmethod
    # def setUpClass(cls) -> None:
    #     cls.remove_db_files()
    #
    # @classmethod
    # def tearDownClass(cls) -> None:
    #     cls.remove_db_files()

    @patch('dm.domain.entities.get_now')
    def setUp(self, mocked_now):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app1 = create_app('test')
        self.app1.config['SECURIZER'] = True
        self.client1 = self.app1.test_client()
        self.app2 = create_app('test')
        self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()

        mocked_now.return_value = dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc)
        with self.app1.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))
            s1 = Server('node1', id='bbbbbbbb-1234-5678-1234-56781234bbb1', port=8000, me=True)
            s2 = Server('node2', id='bbbbbbbb-1234-5678-1234-56781234bbb2', port=8000)
            Route(s2, cost=0)
            db.session.add_all([s1, s2])
            dim = generate_dimension('dimension')
            dim.current = True
            db.session.add(dim)
            db.session.commit()
            self.s1_json = Server.get_current().to_json()
            self.dim_json = dim.to_json()
            self.headers = {"Authorization": f"Bearer {create_access_token('00000000-0000-0000-0000-000000000001')}"}

        with self.app2.app_context():
            db.create_all()
            Locker.set_initial()
            User.set_initial()
            s1 = Server('node1', id='bbbbbbbb-1234-5678-1234-56781234bbb1', port=8000)
            Route(s1, cost=0)
            s2 = Server('node2', id='bbbbbbbb-1234-5678-1234-56781234bbb2', port=8000, me=True)
            db.session.add_all([s1, s2])
            db.session.commit()
            dim = Dimension.from_json(self.dim_json)
            dim.current = True
            db.session.add(dim)
            db.session.commit()

        mocked_now.return_value = dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc)
        with self.app1.app_context():
            soft = Software(id='aaaaaaaa-1234-5678-1234-56781234aaa1', name='test', version='1',
                            filename='file')
            at = ActionTemplate(id='aaaaaaaa-1234-5678-1234-56781234aaa2',
                                name='mkdir', version=1, action_type=ActionType.SHELL, code='mkdir {dir}')
            db.session.add_all([soft, at])
            db.session.commit()
            mocked_now.return_value = dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc)
            ssa = SoftwareServerAssociation(software=soft, server=Server.get_current(), path='/root')
            db.session.add(ssa)
            db.session.commit()
            self.soft_json = soft.to_json()
            self.at_json = at.to_json()

    def tearDown(self) -> None:
        with self.app1.app_context():
            db.session.remove()
            db.drop_all()

        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

    @aioresponses()
    @responses.activate
    def test_check_catalog(self, m):

        set_callbacks([("node1", self.client1), (r"(127\.0\.0\.1|node2)", self.client2)], m)

        with self.app2.app_context():
            self.assertListEqual([('Gate',), ('Server',), ('User',)],
                                 db.session.query(Catalog.entity).order_by(Catalog.entity).all())
            self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Server').last_modified_at)

        with self.app1.app_context():
            self.assertEqual(6, Catalog.query.count())
            self.assertEqual(defaults.INITIAL_DATEMARK, Catalog.query.get('User').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Server').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Gate').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Software').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc),
                             Catalog.query.get('SoftwareServerAssociation').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc),
                             Catalog.query.get('ActionTemplate').last_modified_at)

            resp = self.client1.get(url_for('root.healthcheck'))

        with self.app2.app_context():
            with self.app2.test_request_context('https://node1:5000/api/v1.0/catalog', headers=self.auth.header):
                verify_jwt_in_request()

                s = Server.query.filter_by(_me=False).one()
                update_catalog({s: Response(msg=resp.get_json(), code=resp.status_code)})
                db.session.commit()

        with self.app2.app_context():
            soft = Software.query.get('aaaaaaaa-1234-5678-1234-56781234aaa1')
            self.assertDictEqual(self.soft_json, soft.to_json())
            soft = ActionTemplate.query.get('aaaaaaaa-1234-5678-1234-56781234aaa2')
            self.assertDictEqual(self.at_json, soft.to_json())

            self.assertEqual(6, Catalog.query.count())
            self.assertEqual(defaults.INITIAL_DATEMARK, Catalog.query.get('User').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Software').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 3, tzinfo=dt.timezone.utc),
                             Catalog.query.get('SoftwareServerAssociation').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 2, tzinfo=dt.timezone.utc),
                             Catalog.query.get('ActionTemplate').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Server').last_modified_at)
            self.assertEqual(dt.datetime(2019, 4, 1, tzinfo=dt.timezone.utc),
                             Catalog.query.get('Gate').last_modified_at)



import os
import time
import typing as t
from enum import Enum

import dm.defaults
from dm.domain.entities import Software
from dm.domain.entities.base import EntityReprMixin, UUIDEntityMixin
from dm.utils import typos
from dm.utils.helpers import get_now
from dm.utils.typos import UtcDateTime
from dm.web import db


class Status(Enum):
    WAITING_CHUNKS = 1
    IN_PROGRESS = 2
    COMPLETED = 3
    CHECKSUM_ERROR = 4
    SIZE_ERROR = 5
    CANCELED = 6


class Transfer(db.Model, UUIDEntityMixin, EntityReprMixin):
    __tablename__ = "L_transfer"

    software_id = db.Column(db.ForeignKey('D_software.id'))
    dest_path = db.Column(db.Text, nullable=False)
    num_chunks = db.Column(db.Integer, nullable=False)
    status = db.Column(typos.Enum(Status), nullable=False, default=Status.WAITING_CHUNKS)
    created_on = db.Column(UtcDateTime(timezone=True))
    started_on = db.Column(UtcDateTime(timezone=True))
    ended_on = db.Column(UtcDateTime(timezone=True))
    _filename = db.Column("filename", db.String(256))
    _size = db.Column("size", db.Integer)
    _checksum = db.Column("checksum", db.Text())

    software = db.relationship("Software", uselist=False)

    def __init__(self, software: t.Union[Software, str], dest_path: str, num_chunks: int, status: Status = None,
                 size: int = None, checksum: str = None, created_on=None,
                 **kwargs):
        UUIDEntityMixin.__init__(self, **kwargs)
        if isinstance(software, Software):
            self.software = software
        else:
            self._filename = software
            if size is None:
                ValueError("'size' must be specified when sending a file")
            self._size = size
            if checksum is None:
                ValueError("'checksum' must be specified when sending a file")
            self._checksum = checksum
        self.dest_path = dest_path
        self.num_chunks = num_chunks
        self.status = status or Status.WAITING_CHUNKS
        self.created_on = created_on or get_now()

    def to_json(self):
        json = dict(id=str(self.id), dest_path=self.dest_path,
                    num_chunks=self.num_chunks, status=self.status.name,
                    created_on=self.created_on.strftime(dm.defaults.DATETIME_FORMAT),
                    file=os.path.join(self.dest_path, self.filename))

        if self.software:
            json.update(software_id=str(self.software.id))
        else:
            json.update(size=self._size, checksum=self._checksum)

        if self.started_on:
            json.update(started_on=self.started_on.strftime(dm.defaults.DATETIME_FORMAT))
        if self.ended_on:
            json.update(ended_on=self.ended_on)
        return json

    @property
    def filename(self):
        if self.software:
            return self.software.filename
        else:
            return self._filename

    @property
    def size(self):
        if self.software:
            return self.software.size
        else:
            return self._size

    @property
    def checksum(self):
        if self.software:
            return self.software.checksum
        else:
            return self._checksum

    def wait_transfer(self, timeout=None, refresh_interval: float = None) -> Status:
        timeout = timeout or 300
        refresh_interval = refresh_interval or 1
        start = time.time()
        db.session.refresh(self)
        delta = 0
        while self.status in (Status.IN_PROGRESS, Status.WAITING_CHUNKS) and delta < timeout:
            time.sleep(refresh_interval)
            db.session.refresh(self)
            delta = time.time() - start

        return self.status



_server2id_map = {}

_environ = {}

def set_dict_in_environ(data: dict):
    for k, v in data.items():
        set(k, v)


def set(key, value):
    _environ.update({key: value})


def get(key, default=None):
    return _environ.get(key, default)


def set_server2id_map(data):
    for server in data:
        _server2id_map[server.get('name')] = server.get('id')


def server2id(server):
    return _server2id_map.get(server, None)



import functools
import itertools
import threading
import typing as t

EVENT_EXISTS_ITEM = 2 ** 0
EVENT_UPDATE_ITEM = 2 ** 1


class Event(list):
    def __call__(self, *args, **kwargs):
        for item in self:
            item(*args, **kwargs)


_KT = t.TypeVar('_KT')  # Key type.
_VT = t.TypeVar('_VT')  # Value type.


class Talkback(dict):

    def __init__(self):
        super().__init__()
        self._lock = threading.Lock()
        self._listeners = Event()

    def __setitem__(self, key, value):
        with self._lock:
            self._listeners(key)
            print(f"{key}: {value}")
        super(Talkback, self).__setitem__(key, value)

    def update(self, __m: t.Mapping[_KT, _VT] = None, **kwargs: _VT) -> None:
        with self._lock:
            if isinstance(__m, t.Dict):
                it = __m.items()
            else:
                it = __m or []
            for k, v in itertools.chain(it, kwargs.items()):
                self._listeners(k)
        super().update(__m or {}, **kwargs)

    def _wait_event(self, flag, expected_key, key):
        if expected_key == key:
            flag.set()
            func = None
            for func in self._listeners:
                if flag in func.args:
                    break
            if func:
                self._listeners.remove(func)

    def wait(self, key, event_mask, timeout=None):
        with self._lock:
            if event_mask & EVENT_EXISTS_ITEM:
                if key not in self:
                    flag = threading.Event()
                    func = functools.partial(self._wait_event, flag, key)
                    self._listeners.append(func)
                else:
                    return True
            elif event_mask & EVENT_UPDATE_ITEM:
                flag = threading.Event()
                func = functools.partial(self._wait_event, flag, key)
                self._listeners.append(func)
        r = flag.wait(timeout)
        if not r:
            self._listeners.remove(func)
        return r

    def wait_exists(self, key, timeout=None):
        return self.wait(key, EVENT_EXISTS_ITEM, timeout)

    def wait_update(self, key, timeout=None):
        return self.wait(key, EVENT_UPDATE_ITEM, timeout)
